--
-- Public Lua interface to the Shadow hardware management services. Implements
-- higher level functions in terms of the manager class interface that we bind
-- to the Lua runtime.
--
-- See the "manual/Scripting Reference.pdf" for a complete reference.
--
--


--
-- Load dependencies. The global variable manager is an instance of the manager
-- class that is already bound to a Motion Service runtime.
--
local base = _G
local manager = base.manager
local math = require("math")
local string = require("string")
local table = require("table")

-- Package declaration. All following declarations are in this module table.
local node = {
  _COPYRIGHT = "Copyright 2022 Motion Workshop",
  _NAME = "Node",
  _DESCRIPTION =
    "Public scripting interface to the "
    .. manager.version.product_name
    .. " hardware management services",
  _PRODUCT_NAME = manager.version.product_name,
  _PLATFORM_NAME = manager.version.platform_name,
  _VERSION = manager.version.version,

  -- Default names for some files.
  _initialize_filename = "default/initialize.lua",
  _configuration_filename = "default/configuration.mNode",
  _license_filename = "default/Shadow.mLicense",
  _location_filename = "default/location.mGeocode",
  _ractor_filename = "default/ractor.mRactor",
  _take_filename = "take.mTake"
}

-- For backwards compatibility.
if manager.version.is_bus then
  node._BUS = node._VERSION
end

-- Module local state variables.
local _initialized = false
local _history = {}
local _default_id = "default"
local _service = {}
local _NAME = node._NAME


-- Table of Motion system initialization, configuration, and state functions.
node.system = {}
-- Table of XML string processing related functions. Internal use.
node.xml = {}
-- Table of JSON string processing related functions. Internal use.
node.json = {}
-- Table of User Interface helper functions. Higher level commands that are
-- aggregates of node.* commands.  Internal use.
node.ui = {}


--
-- Begin public function definitions.
--

function node.auto_connect(rescan)
  -- Always close active connections.
  if node.is_connected() then
    node.close()
  end

  -- If we are in full rescan mode, clear out the current configuration.
  if rescan and node.is_configured() then
    node.erase()
  end

  -- Scan for new hardware.
  node.scan()

  -- Start reading from all nodes in the configuration.
  return node.start()
end

function node.close(id)
  -- Only enforce is_connected(id) for single Node close.
  if id then
    if not node.is_connected(id) then
      return false, "failed precondition, \"is_connected(\"" .. id .. "\") == true\""
    end
  end

  -- Stop take.
  if node.is_taking() then
    local result, message = node.stop_take()
    if not result then
      return result, message
    end
  end

  -- Stop reading.
  if node.is_reading(id) then
    local result, message = node.stop(id)
    if not result then
      return result, message
    end
  end

  if id then
    if not manager:close(id) then
      return false, "failed to close connection to " .. _NAME .. " id, \"" .. id .. "\""
    end
  else
    -- Do not call close if the configuration is empty.
    if not node.is_configured() then
      return false, "no configured " .. _NAME .. "s"
    end

    -- Do not call close if the connection list is empty.
    if not node.connected() then
      return false, "no connected " .. _NAME .. "s"
    end

    if not manager:close("") then
      return false, "failed to close at least one open connection"
    end
  end

  return true, ""
end

function node.configuration()
  return manager:configuration()
end

function node.connect(id)
  if not node.is_configured(id) then
    if id then
      return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
    else
      return false, "failed precondition \"is_configured() == true\""
    end
  end

  if node.is_connected(id) then
    if id then
      return false, "failed precondition \"is_connected(\"" .. id .. "\") == false\""
    else
      return false, "failed precondition \"is_connected() == false\""
    end
  end

  if id then
    if not manager:connect(id) then
      return false, "failed to connect to " .. _NAME .. " id, \"" .. id .. "\""
    end
  else
    if not manager:connect("") then
      return false, "failed to connect to at least one " .. _NAME .. " in configuration"
    end
  end

  return true, ""
end

function node.connected()
  return manager:connected()
end

function node.enumerate(name)
  if node.is_connected() then
    return nil
  end
  if node.connected() then
    return nil
  end

  if not name then
    name = ""
  end

  return manager:enumerate(name)
end

function node.erase(id)
  if not node.is_configured(id) then
    if id then
      return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
    else
      return false, "failed precondition \"is_configured() == true\""
    end
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if id then
    if not manager:erase(id) then
      return false, "failed to erase " .. _NAME .. " configuration with id \"" .. id .. "\""
    end
  else
    if not manager:erase("") then
      return false, "failed to erase " .. _NAME .. " configuration"
    end
  end

  return true, ""
end

function node.export(filename, take_filename, option, file_type, mode)
  if not filename then
    filename = ""
  end

  if not take_filename then
    local result
    result, take_filename = node.get_take_path()

    if result then
      take_filename = node.system.file_canonical(take_filename)
    else
      take_filename = ""
    end
  end

  _history["export"] = filename

  if not node.system.file_exists(take_filename) then
    return false, "failed precondition \"system.file_exists(\"" .. take_filename .. "\") == true\""
  end

  if not node.system.is_filename(filename) then
    return false, "failed precondition \"system.is_filename(\"" .. filename .. "\") == true\""
  end

  if not file_type or string.len(file_type) == 0 then
    _, _, file_type = string.find(filename, "%.([^%.]+)$")
  end

  if not mode then
    mode = 0
  end

  if not file_type then
    file_type = node.system.export_default_type(mode)
  end

  local valid_type = node.system.export_type(mode, file_type)
  if not valid_type then
    return false, "invalid export type \"" .. base.tostring(file_type) .. "\" for file \"" .. filename .. "\""
  end

  -- Append the file extension if it is not already there.
  if not string.find(filename, "%." .. file_type .. "$") then
    filename = filename .. "." .. file_type
  end

  -- Create a list of the default options for this file type. Every type inherits
  -- the same base set of options.
  if not option then
    if valid_type["option"] and (string.len(valid_type["option"]) > 0) then
      option = valid_type["option"]
    else
      option = ""
    end
  end

  _history["export_type"] = file_type
  _history["export_option"] = option

  if not manager:export(filename, take_filename, option, file_type, mode) then
    return false, "failed to export take to file \"" .. filename .. "\""
  end

  filename = node.system.file_canonical(filename)
  _history["export"] = filename

  return true, filename
end

function node.export_stream(filename, take_filename, option, file_type)
  return node.export(filename, take_filename, option, file_type, 2)
end

function node.get_data_path()
  local pref = node.system.get_preference()
  if pref then
    return true, pref.data_path
  else
    return false, ""
  end
end

function node.get_node(key, value)
  if not key then
    return nil
  end

  if key == "id" and _default_id == value then
    return node.get_node_default()
  end

  local list = node.configuration()
  if not list then
    return nil
  end

  local item
  for _,item in base.pairs(list) do
    if value == item[key] then
      return base.copy_usertype(item)
    end
  end

  return nil
end

function node.get_node_ancestor(id)
  if not id then
    id = ""
  end

  if not node.is_id(id) then
    return nil
  end

  if not node.is_configured(id) then
    return nil
  end

  return manager:get_node_ancestor(id)
end

function node.get_node_by_id(id)
  return node.get_node("id", id)
end

function node.get_node_by_key(key)
  return node.get_node("key", key)
end

function node.get_node_default()
  return manager:get_node_default()
end

function node.get_node_visit_order()
  if not node.is_configured() then
    return nil
  end

  return manager:get_node_visit_order()
end

function node.get_version()
  return manager.version
end

function node.get_take()
  return manager:get_take()
end

function node.get_take_path()
  if node.have_take() then
    local take = node.get_take()
    if take then
      if string.len(take.working) > 0 and string.len(take.info_path) > 0 then
        return true, take.working .. "/" .. take.info_path
      else
        local result, path = node.get_data_path()
        if result then
          return true, path .. "/" .. take.name .. "/" .. node._take_filename
        end
      end
    end
  end

  return false, ""
end

function node.have_take()
  return manager:have_take()
end

function node.insert(id)
  _history["insert"] = id

  if not node.is_id(id) then
    if not id then
      id = ""
    end
    return false, "failed precondition \"is_id(\"" .. id .. "\") == true\""
  end
  local result, id = node.system.unique_id(id)
  if not result then
    return false, "failed precondition \"system.unique_id(\"" .. id .. "\") == true\""
  end
  if node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == false\""
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not manager:insert(id, false, "", "") then
    return false, "failed to insert " .. _NAME .. " configuration with id \"" .. id .. "\""
  end

  return true, id
end

function node.insert_node(obj)
  if not obj then
    return nil
  end

  if not node.is_id(obj.id) then
    return nil
  end

  if node.is_configured(obj.id) then
    return nil
  end

  if node.is_connected() then
    return nil
  end

  if node.connected() then
    return nil
  end

  return manager:insert_node(obj, true)
end

function node.is_configured(id)
  if not id then
    id = ""
  end

  if id == _default_id then
    return true
  end

  return manager:is_configured(id)
end

function node.is_connected(id)
  if not id then
    id = ""
  end

  return manager:is_connected(id)
end

function node.is_id(id)
  if not id then
    return false
  end

  if string.len(id) == 0 then
    return false
  end

  return manager:is_id(id)
end

function node.is_reading(id)
  if not id then
    id = ""
  end

  return manager:is_reading(id)
end

function node.is_taking()
  return manager:is_taking()
end

function node.is_tracking()
  return manager:is_tracking()
end

function node.load_configuration(filename)
  if not filename then
    filename = ""
  end

  _history["load_configuration"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not manager:load_configuration(filename) then
    return false, "failed to load configuration file \"" .. filename .. "\""
  end

  _history["load_configuration"] = node.system.file_canonical(filename)

  return true, ""
end

function node.load_license(filename)
  if not filename then
    filename = ""
  end

  _history["load_license"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end

  if not manager:load_license(filename) then
    return false, "failed to load license file \"" .. filename .. "\""
  end

  _history["load_license"] = node.system.file_canonical(filename)

  return true, ""
end

function node.load_location(filename)
  if not filename then
    filename = ""
  end

  _history["load_location"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not manager:load_location(filename) then
    return false, "failed to load location file \"" .. filename .. "\""
  end

  _history["load_location"] = node.system.file_canonical(filename)

  return true, ""
end

function node.load_ractor(filename)
  if not filename then
    filename = ""
  end

  _history["load_ractor"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not manager:load_ractor(filename) then
    return false, "failed to load ractor definition file \"" .. filename .. "\""
  end

  _history["load_ractor"] = node.system.file_canonical(filename)

  return true, ""
end

function node.load_script(filename)
  if not filename then
    filename = ""
  end

  _history["load_script"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end

  if not manager:load_script(filename) then
    return false, "failed to load script file \"" .. filename .. "\""
  end

  _history["load_script"] = node.system.file_canonical(filename)

  return true, ""
end

function node.load_take(filename)
  if not filename then
    filename = ""
  end

  _history["load_take"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not manager:load_take(filename) then
    return false, "failed to load take file \"" .. filename .. "\""
  end

  _history["load_take"] = node.system.file_canonical(filename)

  return true, ""
end

function node.load_take_source(filename)
  if filename then
    _history["load_take_source"] = filename

    local result, message = node.load_take(filename)
    if not result then
      return result, message
    end

    _history["load_take_source"] = node.system.file_canonical(filename)
  end

  if not node.have_take() then
    return false, "failed precondition \"have_take() == true\""
  end
  if not node.is_configured() then
    return false, "failed precondition \"is_configured() == true\""
  end
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end
  if not node.system.configuration_matches_take() then
    return false, "failed precondition \"system.configuration_matches_take() == true\""
  end

  if not manager:load_take_source() then
    return false, "failed to load take as " .. _NAME .. " source"
  end

  return true, ""
end

function node.merge_configuration(source, tree_doc)
  if node.is_connected() then
    return false, "failed precondition, \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not source then
    source = ""
  end
  if not tree_doc then
    tree_doc = ""
  end

  if manager:merge_configuration(source, tree_doc) then
    return true, ""
  else
    return false, "failed to merge enumerated tree into configuration"
  end
end

function node.num_active(list, active)
  if not list then
    return 0
  end

  if not active then
    return #list
  end

  local num_node = 0

  local item
  for _,item in base.pairs(list) do
    if not item.bus and item.active then
      num_node = num_node + 1
    end
  end

  return num_node
end

function node.num_configured(active)
  return node.num_active(node.configuration(), active)
end

function node.num_connected(active)
  return node.num_active(node.connected(), active)
end

function node.num_reading(active)
  return node.num_active(node.reading(), active)
end

function node.open(filename)
  if not filename then
    filename = ""
  end

  _history["open"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end

  -- Automatically close active connections when we call file open.
  local reconnect = node.is_connected() and node.close()

  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not manager:open(filename) then
    return false, "failed to open file \"" .. filename .. "\""
  end

  if reconnect then
    node.start()
  end

  _history["open"] = node.system.file_canonical(filename)

  return true, ""
end

function node.parent(id, parent_id)
  if not id then
    return false, "missing required id parameter"
  end

  if not parent_id then
    return false, "missing required parent_id parameter"
  end

  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end

  if not node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
  end

  if not node.is_configured(parent_id) then
    return false, "failed precondition \"is_configured(\"" .. parent_id .. "\") == true\""
  end

  if not manager:parent(id, parent_id) then
    return false, "failed to parent " .. _NAME .. " id, \"" .. id .. "\" to \"" .. parent_id .. "\""
  end

  return true, ""
end

function node.reading()
  return manager:reading()
end

function node.save_configuration(filename)
  if not filename then
    filename = ""
  end

  _history["save_configuration"] = filename

  if not node.system.is_filename(filename) then
    return false, "failed precondition \"system.is_filename(\"" .. filename ..
                  "\") == true\""
  end
  if not node.is_configured() then
    return false, "failed precondition \"is_configured() == true\""
  end

  if not manager:save_configuration(filename) then
    return false, "failed to save configuration file \"" .. filename .. "\""
  end

  filename = node.system.file_canonical(filename)
  _history["save_configuration"] = filename

  return true, filename
end

function node.scan(name)
  if node.is_connected() then
    return false, "failed precondition, \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  -- Run all device enumeration routines. Optionally only scan for devices of a
  -- specific class name, like "usb" or "bus".
  local enumeration = node.enumerate(name)
  if not enumeration then
    return false, "no devices found"
  end

  -- Grab the current configuration list.
  local configured = node.configuration()
  local in_configuration = {}

  -- Create a table of existing hardware sources. Key them by the hardware
  -- source.
  --   in_configuration["bus://Shadow"] = true
  --   in_configuration["bus://Shadow/Hips"] = true
  --   in_configuration["usb://100-123456"] = true
  if configured then
    local item
    for _,item in base.pairs(configured) do
      local source = item.source

      -- Trim off the option string.
      local stype, suri = string.match(source, "([^:]+)://([^:]+)")
      if stype and suri then
        source = stype .. "://" .. suri
      end

      in_configuration[source] = true
    end
  end

  local num_device = 0
  local num_device_existing = 0
  local item_with_tree = {}

  -- Iterate through the enumeration result and add the devices to the
  -- configuration.
  local item
  for _,item in base.pairs(enumeration) do
    if not in_configuration[item.source] then
      if manager:insert_node(item, false) ~= nil then
        num_device = num_device + 1

        -- If the node has a configuration tree string, a companion definition,
        -- save a reference so we can run the tree merge later.
        if item.tree_doc and item.tree_doc ~= '' then
          table.insert(item_with_tree, item)
        end
      end
    else
      num_device_existing = num_device_existing + 1
    end
  end

  -- Group a bunch of configuration updates from insert into one call.
  if num_device > 0 then
    manager:insert_node(nil, true)
  end

  -- Search for enumerated devices that have an associated tree structure. Merge
  -- the tree structure and flat device list.
  for _,item in base.pairs(item_with_tree) do
    manager:merge_configuration(item.source, item.tree_doc)
  end

  if num_device == 0 then
    if num_device_existing == 0 then
      return false, "no devices found"
    else
      return false, "no additional devices found"
    end
  end

  return true, ""
end

function node.set_configuration(key, value, id, ignore_children)
  if not node.is_configured(id) then
    if id then
      return false, "failed precondition, \"is_configured(\"" .. id .. "\") == true\""
    else
      return false, "failed precondition, \"is_configured() == true\""
    end
  end

  local result = false

  if id then
    -- If the id parameter is set, operate on a single node.
    local item = node.get_node_by_id(id)
    if item then
      item[key] = value
      result = true
    end

    -- Propagate change to any node that lists this as its parent.
    if not ignore_children then
      local list = node.configuration()
      if list then
        for _,item in base.pairs(list) do
          if id == item.parent then
            item[key] = value
          end
        end
      end
    end
  else
    local list = node.configuration()
    if list then
      local item
      for _,item in base.pairs(list) do
        item[key] = value
        result = true
      end
    end
  end

  return result, ""
end

function node.set_gain(value, id)
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  return node.set_configuration("gain", value, id)
end

function node.set_gain_sensor(value, id)
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  return node.set_configuration("gain_sensor", value, id)
end

function node.set_gselect(value, id)
  if node.is_connected(id) then
    if id then
      return false, "failed precondition \"is_connected(\"" .. id .. "\") == false\""
    else
      return false, "failed precondition \"is_connected() == false\""
    end
  end

  return node.set_configuration("gselect", value, id)
end

function node.set_hierarchy(id, fn, rx, ry, rz, tx, ty, tz)
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  if not node.is_configured(id) then
    if id then
      return false, "failed precondition, \"is_configured(\"" .. id .. "\") == true\""
    else
      return false, "failed precondition, \"is_configured() == true\""
    end
  end

  if base.type(manager[fn]) ~= "function" then
    return false, "hierarchy interface not implemented"
  end

  if id then
    if tx == nil and manager[fn](manager, id, rx, ry, rz) then
    elseif manager[fn](manager, id, rx, ry, rz, tx, ty, tz) then
    else
      return false, "failed to update hierarchy for " .. _NAME .. " id, \"" .. id .. "\""
    end
  else
    local list = node.configuration()
    if not list then
      return false, "failed precondition, \"configuration() ~= nil\""
    end

    local item
    for _,item in base.pairs(list) do
      local id = item.id
      if tx == nil and manager[fn](manager, id, rx, ry, rz) then
      elseif manager[fn](manager, id, rx, ry, rz, tx, ty, tz) then
      else
        return false, "failed to update hierarchy for " .. _NAME .. " id, \"" .. id .. "\""
      end
    end

  end

  return true, ""
end

function node.set_marker(id, x, y, z, tx, ty, tz)
  if not(tx) or not(ty) or not(tz) then
    tx = 0
    ty = 0
    tz = 0
  end

  return node.set_hierarchy(id, "set_node_marker", x, y, z, tx, ty, tz)
end

function node.set_name(value, id)
  if node.is_connected(id) then
    if id then
      return false, "failed precondition \"is_connected(\"" .. id .. "\") == false\""
    else
      return false, "failed precondition \"is_connected() == false\""
    end
  end

  return node.set_configuration("name", value, id, true)
end

function node.set_node_pose(id, x, y, z)
  return node.set_hierarchy(id, "set_node_pose", x, y, z)
end

function node.set_pipeline_select(value, id)
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  return node.set_configuration("pipeline_select", value, id)
end

function node.set_pose(id)
  if not node.is_reading() then
    return false, "failed precondition, \"is_reading() == true\""
  end
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  if id then
    if not manager:set_pose(id) then
      return false, "failed to set zero pose for \"" .. id .. "\" subtree"
    end
  else
    if not manager:set_pose("") then
      return false, "failed to set zero pose for at least one configured " .. _NAME
    end
  end

  return true, ""
end

function node.set_pose_marker(id)
  if not node.is_reading() then
    return false, "failed precondition, \"is_reading() == true\""
  end
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  if id then
    if not manager:set_pose_marker(id) then
      return false, "failed to create rest pose orientation marker for \"" .. id .. "\" subtree"
    end
  else
    if not manager:set_pose_marker("") then
      return false, "failed to create rest pose orientation marker for at least one configured " .. _NAME
    end
  end

  return true, ""
end

function node.set_rotate(id, x, y, z)
  return node.set_hierarchy(id, "set_node_rotate", x, y, z)
end

function node.set_source(value, id)
  if node.is_connected(id) then
    if id then
      return false, "failed precondition \"is_connected(\"" .. id .. "\") == false\""
    else
      return false, "failed precondition \"is_connected() == false\""
    end
  end

  return node.set_configuration("source", value, id, true)
end

function node.set_time_step(value, id)
  if node.is_connected(id) then
    if id then
      return false, "failed precondition \"is_connected(\"" .. id .. "\") == false\""
    else
      return false, "failed precondition \"is_connected() == false\""
    end
  end

  return node.set_configuration("time_step", value, id)
end

function node.set_track_gyroscope_bias(value, id)
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  return node.set_configuration("track_gyroscope_bias", value, id)
end

function node.set_translate(id, x, y, z)
  return node.set_hierarchy(id, "set_node_translate", x, y, z)
end

function node.start(id)
  if node.is_reading(id) then
    if id then
      return false, "failed precondition \"is_reading(\"" .. id .. "\") == false\""
    else
      return false, "failed precondition \"is_reading() == false\""
    end
  end

  -- Connect.
  if not node.is_connected(id) then
    local result, message = node.connect(id)
    if not result then
      return result, message
    end
  end

  if id then
    if not manager:start(id) then
      return false, "failed to start reading from " .. _NAME .. " id, \"" .. id .. "\""
    end
  else
    if not manager:start("") then
      return false, "failed to start reading from at least one " .. _NAME .. " in configuration"
    end
  end

  return true, ""
end

function node.start_take(name, description)
  if node.is_taking() then
    return false, "failed precondition, \"is_taking() == false\""
  end

  -- Start reading.
  if not node.is_reading() then
    local result, message = node.start()
    if not result then
      return result, message
    end
  end

  -- Create default parameter values.
  if not name then
    name = ""
  end
  if not description then
    description = ""
  end

  -- Start take.
  if not manager:start_take(name, description) then
    return false, "failed to start take"
  end

  return true, ""
end

function node.stop(id)
  -- Only enforce is_reading(id) for single Node stop.
  if id then
    if not node.is_reading(id) then
      return false, "failed precondition, \"is_reading(\"" .. id .. "\") == true\""
    end
  end

  -- Stop take.
  if node.is_taking() then
    local result, message = node.stop_take()
    if not result then
      return result, message
    end
  end

  -- Stop reading.
  if id then
    if not manager:stop(id) then
      return false, "failed to stop reading from " .. _NAME .. " id, \"" .. id .. "\""
    end
  else
    -- Do not call close if the configuration is empty.
    if not node.is_configured() then
      return false, "no configured " .. _NAME .. "s"
    end

    -- Do not call stop if the reading list is empty.
    if not node.reading() then
      return false, "no reading " .. _NAME .. "s"
    end

    if not manager:stop("") then
      return false, "failed to stop reading from at least one " .. _NAME .. " in configuration"
    end
  end

  return true, ""
end

function node.stop_take()
  if not manager:is_taking() then
    return false, "failed precondition, \"is_taking() == true\""
  end

  -- Stop take.
  if not manager:stop_take() then
    return false, "failed to stop take"
  end

  return true, ""
end

function node.track_take(filename)
  if not filename then
    filename = ""
  end

  _history["track_take"] = filename

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end

  if not manager:track_take(filename) then
    return false, "failed to track take from file \"" .. filename .. "\""
  end

  _history["track_take"] = node.system.file_canonical(filename)

  return true, ""
end

function node.write_configuration(id)
  if not id then
    id = ""
  end

  if not node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
  end

  if node.is_connected(id) then
    return false, "failed precondition \"is_connected(\"" .. id .. "\") == false\""
  end

  if not manager:write_configuration(id) then
    return false, "failed to write configuration to " .. _NAME .. " id, \"" .. id .. "\""
  end

  return true, ""
end


function node.system.authorize_device(uuid, token)
  if not uuid then
    uuid = ""
  end

  if not token then
    token = ""
  end

  if not manager:authorize_device(uuid, token) then
    return false, "failed to authorize device"
  end

  return true, ""
end

function node.system.calibrate_from_take(id, calibrate_accel, calibrate_gyro)
  if not node.have_take() then
    return false, "failed precondition \"have_take() == true\""
  end
  if not node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
  end
  if not node.system.configuration_matches_take() then
    return false, "failed precondition \"system.configuration_matches_take() == true\""
  end

  -- Only calibrate the accelerometers if we specifically ask for it.
  if not calibrate_accel then
    calibrate_accel = false
  end

  -- Gyroscope only mode. Do not rotate, stationary sensor data.
  if not calibrate_gyro then
    calibrate_gyro = false
  end

  if not manager:calibrate_from_take(id, calibrate_accel, calibrate_gyro) then
    return false, "failed to generate " .. _NAME .. " calibration from take data"
  end

  return true, ""
end

function node.system.configuration_matches_take()
  if not node.have_take() then
    return false, "failed precondition \"have_take() == true\""
  end
  if not node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
  end

  if not manager:configuration_matches_take() then
    return false, "current configuration does not match current take"
  end

  return true, ""
end

function node.system.export_default_type(mode)
  if not _history["export_type"] then
    local file_type = node.system.get_default("export_type")
    if not file_type then
      file_type = "fbx"
    end

    return file_type
  else
    return _history["export_type"]
  end
end

function node.system.export_stream_type(file_type)
  return node.system.export_type(2, file_type)
end

function node.system.export_type(mode, file_type)
  if not mode then
    mode = 0
  end

  if not file_type then
    file_type = ""
  end

  local list = manager:get_export_type_list(mode, file_type)
  if not list then
    return nil
  end

  if file_type then
    local item
    for _,item in base.pairs(list) do
      if item.id == file_type then
        return base.copy_usertype(item)
      end
    end
  end

  return list
end

function node.system.file_canonical(filename)
  if not filename then
    filename = ""
  end

  return manager:file_canonical(filename)
end

function node.system.file_exists(filename)
  if not filename then
    filename = ""
  end

  if not manager:file_exists(filename) then
    return false
  end

  return true
end

function node.system.geocode_address(address, name)
  if not address then
    address = ""
  end

  if not name then
    name = ""
  end

  if not manager:geocode_address(address, name) then
    return false, "failed to geocode address \"" .. address .. "\""
  end

  return true, ""
end

function node.system.get_default(name)
  if not name then
    name = ""
  end

  return manager:get_default(name)
end

function node.system.get_history()
  return _history
end

function node.system.get_local_mode()
  local mode = node.system.get_default("local_mode")

  if mode then
    mode = mode + 0
  end

  local key, value
  for key,value in base.pairs(node.system.local_mode()) do
    if key == mode then
      return value
    end
  end

  return nil
end

function node.system.get_location_list()
  return manager:get_location_list()
end

function node.system.get_preference()
  return manager:get_preference()
end

function node.system.get_ractor_list()
  return manager:get_ractor_list()
end

function node.system.get_service()
  return _service
end

function node.system.get_settings(id)
  if not id then
    id = ""
  end

  if not node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
  end

  if node.is_reading(id) then
    return false, "failed precondition \"is_reading(\"" .. id .. "\") == false\""
  end

  if not manager:get_settings(id) then
    return false, "failed to read settings for " .. _NAME .. " id, \"" .. id .. "\""
  end

  return true, ""
end

function node.system.get_system_conf_list()
  return manager:get_system_conf_list()
end

function node.system.get_system_db_list()
  return manager:get_system_db_list("")
end

function node.system.get_system_info()
  return manager:get_system_info()
end

function node.system.get_wifi_list()
  return manager:get_wifi_list()
end

function node.system.get_zeroconf_list()
  return manager:get_zeroconf_list()
end

function node.system.initialize()
  if false == _initialized then
    local filename = node._initialize_filename

    if not node.system.file_exists(filename) then
      return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
    end

    if node.load_script(filename) then
      _initialized = true
    else
      return false, "failed to load initialization file, \"" .. filename .. "\""
    end
  end

  return true, ""
end

function node.system.initialize_session()
  if not manager:initialize_session() then
    return false, "failed to initialize session settings"
  end

  _history = {}

  return true, ""
end

function node.system.is_authorized(uuid)
  if not uuid then
    uuid = ""
  end

  local configured_node = node.get_node("uuid", uuid)
  if configured_node then
    if configured_node.auth > 0 and configured_node.auth < 5 then
      return false, "device not authorized"
    end
  else
    return false, "failed to find device with uuid, \"" .. uuid .. "\""
  end

  return true, ""
end

function node.system.is_filename(filename)
  if not filename then
    filename = ""
  end

  return manager:is_filename(filename)
end

function node.system.is_initialized()
  return true == _initialized
end

function node.system.load_plugin(name)
  if node.is_connected() then
    return false, "failed precondition, \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not name then
    name = ""
  end

  if not manager:load_plugin(name) then
    if string.len(name) > 0 then
      return false, "failed load plugin named \"" .. name .. "\""
    else
      return false, "failed to load at least one plugin"
    end
  end

  return true, ""
end

function node.system.local_mode()
  local result = {
    [0] = "NONE",
    [1] = "WORLD",
    [2] = "SENSOR",
    [3] = "WORLD_HEADING",
    [4] = "HIERARCHY",
    [5] = "MARKER"
  }
  return result
end

function node.system.location(latitude, longitude, elevation, address, name)
  if latitude <= -90 or latitude >= 90 then
    return false, "failed precondition \"-90 < latitude < 90\""
  end

  if longitude <= -180 or longitude >= 180 then
    return false, "failed precondition \"-180 < longitude < 180\""
  end

  if elevation < -1000 or elevation > 600000 then
    return false, "failed precondition \"-1000 <= elevation <= 600000\""
  end

  if not address then
    address = ""
  end

  if not name then
    name = ""
  end

  if not manager:location(latitude, longitude, elevation, address, name) then
    return false, "failed to set location, invalid parameter"
  end

  return true, ""
end

function node.system.log(filename)
  if not filename then
    filename = ""
  end

  if not node.system.is_filename(filename) then
    return false, "failed precondition \"system.is_filename(\"" .. filename .. "\") == true\""
  end

  if node.system.is_initialized() then
    return false, "failed precondition \"system.is_initialized() == false\""
  end

  if not manager:log(filename) then
    return false, "failed to set log file \"" .. filename .. "\""
  end

  return true, ""
end

function node.system.open_database(filename)
  if not filename then
    filename = ""
  end

  if not node.system.file_exists(filename) then
    return false, "failed precondition \"system.file_exists(\"" .. filename .. "\") == true\""
  end

  if node.system.is_initialized() then
    return false, "failed precondition \"system.is_initialized() == false\""
  end

  if not manager:open_database(filename) then
    return false, "failed to open database file \"" .. filename .. "\""
  end

  -- Load up export type preference.
  local file_type = node.system.get_default("export_type")
  if file_type then
    local valid_type = node.system.export_type(0, file_type)
    if valid_type then
      _history["export"] = "take." .. file_type
      _history["export_type"] = file_type
    end
  end

  return true, ""
end

function node.system.print(...)
  local result = ""

  local value
  for _,value in base.ipairs({...}) do
    result = result .. base.tostring(value)
  end

  result = result .. "\n"

  manager:print(result)
end

function node.system.quit()
  return manager:quit()
end

function node.system.ractor(name, height, arm, leg)
  if not name then
    name = ""
  end

  if not height then
    height = -1
  end

  if not arm then
    arm = -1
  end

  if not leg then
    leg = -1
  end

  if not manager:ractor(name, height, arm, leg) then
    return false, "failed to configure ractor from measurements, invalid parameter"
  end

  return true, ""
end

function node.system.republish_services()
  return manager:republish_services()
end

function node.system.reset_ractor()
  return manager:reset_ractor()
end

function node.system.reset_wifi()
  return manager:reset_wifi()
end

function node.system.restart()
  return manager:restart()
end

function node.system.save_initialization()
  local filename = node._initialize_filename
  local script_filename = "api/initialize.lua"

  if not node.system.is_filename(filename) then
    return false, "failed precondition \"system.is_filename(\"" .. filename .. "\") == true\""
  end
  if not node.system.file_exists(script_filename) then
    return false, "failed precondition \"system.file_exists(\"" .. script_filename .. "\") == true\""
  end

  if not manager:save_initialization(filename, script_filename) then
    return false, "failed to save initialization to file, \"" .. filename .. "\""
  end

  -- If the default location file exists at this point, write the current
  -- location out to it.
  if node.system.file_exists(node._location_filename) then
    local result, message = node.system.save_location()
    if not result then
      return false, message
    end
  end

  return node.system.initialize()
end

function node.system.save_location(filename)
  if not filename then
    filename = node._location_filename
  end

  if not node.system.is_filename(filename) then
    return false, "failed precondition \"system.is_filename(\"" .. filename .. "\") == true\""
  end

  if not manager:save_location(filename) then
    return false, "failed to save current location to file, \"" .. filename .. "\""
  end

  filename = node.system.file_canonical(filename)

  return true, filename
end

function node.system.save_ractor(filename)
  if not filename then
    filename = node._ractor_filename
  end

  if not node.system.is_filename(filename) then
    return false, "failed precondition \"system.is_filename(\"" .. filename .. "\") == true\""
  end

  if not manager:save_ractor(filename) then
    return false, "failed to save current ractor definition to file, \"" .. filename .. "\""
  end

  filename = node.system.file_canonical(filename)

  return true, filename
end

function node.system.set_data_path(path)
  if not path then
    path = ""
  end
  if node.system.is_initialized() then
    return false, "failed precondition \"system.is_initialized() == false\""
  end

  if not manager:set_data_path(path) then
    return false, "failed to set data path \"" .. path .. "\""
  end

  return true, ""
end

function node.system.set_date_time(timestamp)
  if not timestamp then
    timestamp = ""
  end

  return manager:set_date_time(timestamp)
end

function node.system.set_default(name, value)
  if not name then
    name = ""
  end
  if nil == value then
    value = ""
  else
    value = base.tostring(value)
  end

  return manager:set_default(name, value)
end

function node.system.set_local_mode(value)
  if not value then
    value = 1
  end

  if base.type(value) == base.type("") then
    value = string.upper(value)

    for i,name in base.pairs(node.system.local_mode()) do
      if name == value then
        value = i
        break
      end
    end

  elseif value < 0 then
    value = 0
  elseif value > 5 then
    value = 5
  end

  return manager:set_local_mode(value)
end

function node.system.set_search_path(path)
  if not path then
    path = ""
  end
  if node.system.is_initialized() then
    return false, "failed precondition \"system.is_initialized() == false\""
  end

  if not manager:set_search_path(path) then
    return false, "failed to set search path \"" .. path .. "\""
  end

  return true, ""
end

function node.system.set_timecode(value)
  if not value then
    value = ""
  end

  if manager:set_timecode(value) then
    return true, ""
  else
    return false, "failed to set system timecode to \"" .. value .. "\""
  end
end

function node.system.set_wifi(hwid, essid, password)
  if not hwid then
    hwid = "Any"
  end
  if not essid then
    essid = ""
  end
  if not password then
    password = ""
  end

  if manager:set_wifi(hwid, essid, password) then
    return true, ""
  else
    return false, "failed to select wifi network"
  end
end

function node.system.shutdown(id)
  if not(id) then
    id = ""
  end

  if manager:shutdown(id) then
    return true, ""
  else
    if node.is_id(id) then
      return false, "failed to send shutdown command to " .. _NAME .. " id, \"" .. id .. "\""
    else
      return false, "failed to send shutdown command"
    end
  end
end

function node.system.sleep(seconds)
  if not(seconds) then
    seconds = 0
  end

  return manager:sleep(seconds)
end

function node.system.start_services(service)
  if node.system.is_initialized() then
    return false, "failed precondition \"system.is_initialized() == false\""
  end

  local num_running = 0
  local message = ""

  local name, port
  for name, port in base.pairs(service) do
    -- Skip service names that are already in our table.
    if not _service[name] then
      local obj = manager:service(name, port)
      if not obj then
        -- Only describes the last failed service.
        message = "failed to start service \"" .. name .. "\" on port " .. base.tostring(port)
      else
        _service[name] = obj
        num_running = num_running + 1
      end
    else
      num_running = num_running + 1
    end
  end

  if num_running > 0 then
    return true, ""
  end

  return false, message
end

function node.system.trigger_state_change(code, data)
  if not code then
    code = 1
  end
  if not data then
    data = ""
  end

  return manager:trigger_state_change(code, data)
end

function node.system.unique_id(id)
  if not node.is_id(id) then
    return false, "failed precondition \"is_id(\"" .. id .. "\") == true\""
  end

  if not node.is_configured(id) then
    return true, id
  end

  id = manager:unique_id(id)
  return (string.len(id) > 0), id
end

function node.system.unload_plugin(name)
  if node.is_connected() then
    return false, "failed precondition, \"is_connected() == false\""
  end
  if node.connected() then
    return false, "failed precondition \"connected() == nil\""
  end

  if not name then
    name = ""
  end

  if not manager:unload_plugin(name) then
    return false, "failed to unload at least one plugin"
  end

  return true, ""
end

function node.system.command_to_string(format, command, ...)
  local fn = node[command]

  -- Split commands of the form "table.command" and bind fn to the
  -- node[table][command] entry.
  local command1, command2
  _, _, command1, command2 = string.find(command, "^([^%.]+)%.([^%.]+)$")
  if command1 and command2 then
    fn = node[command1][command2]
  end

  if not fn or base.type(fn) ~= "function" then
    base.error("function not found in node interface, \"node." .. command .. "\"")
  end

  return format(fn(...))
end

--
-- End public function definitions.
--

--
-- Begin internal function definitions. Not part of the documented API and
-- subject to change with no backwards compatibility.
--

function node.xml.encode_attribute(value)
  local result = node.xml.encode_string(value)
  result = string.gsub(result, "\"", "&quot;")
  return result
end

function node.xml.encode_string(value)
  local result = base.tostring(value)
  result = string.gsub(result, "&", "&amp;")
  result = string.gsub(result, "<", "&lt;")
  result = string.gsub(result, ">", "&gt;")
  result = string.gsub(result, "\\", "\\\\")
  return result
end

function node.xml.result_to_xml(...)
  -- Make a table of type conversions for XML output.
  local xml_type = {
    ["string"] = "string",
    ["number"] = "number",
    ["boolean"] = "boolean"
  }

  local result = ""

  local value
  for _,value in base.ipairs({...}) do
    local value_type = xml_type[base.type(value)]
    if not value_type then
      -- Unknown XML type, convert to string.
      value = base.tostring(value)
      value_type = "string"
    end

    result = result .. "<param><value><" .. value_type
    if value ~= nil then
      if base.type(value) == "boolean" then
        if true == value then
          value = 1
        else
          value = 0
        end
      end

      result =
        result .. ">" .. node.xml.encode_string(value) ..
        "</" .. value_type .. ">"
      else
        result = result .. "/>"
    end
    result = result .. "</value></param>"
  end

  return result
end

function node.json.encode_string(value)
  local result = base.tostring(value)
  result = string.gsub(result, "\\", "\\\\")
  result = string.gsub(result, '"', '\\"')
  result = string.gsub(result, '\t', '\\t')
  result = string.gsub(result, '\n', '\\n')
  return result
end

function node.json.is_list(value)
  -- By our convention, a list is any table with ascending integer keys
  -- starting at 1.
  if base.type({}) ~= base.type(value) then
    return false
  end

  if 0 == #value then
    return false
  end

  local index = 1
  for k,v in base.pairs(value) do
    if base.type(0) ~= base.type(k) then
      return false
    end

    if k ~= index then
      return false
    end

    index = index + 1
  end

  return true
end

function node.json.encode(value, is_list)
  if nil == value then
    return "null"
  end

  local value_type = base.type(value)

  -- Regular strings.
  if value_type == "string" then
    return '"' .. node.json.encode_string(value) .. '"'
  end

  -- Numbers and booleans.
  if value_type == "number" or value_type == "boolean" then
    return base.tostring(value)
  end

  -- Tables.
  if value_type == "table" then
    if is_list == nil then
      is_list = node.json.is_list(value)
    end

    local kv = {}

    local k, v
    for k,v in base.pairs(value) do
      if is_list then
        table.insert(kv, node.json.encode(v))
      else
        table.insert(
          kv, '"' .. node.json.encode_string(k) .. '": ' ..
          node.json.encode(v, ('node' == k) or ('items' == k) or nil))
      end
    end

    if is_list then
      return "[" .. table.concat(kv, ", ") .. "]"
    else
      return "{" .. table.concat(kv, ", ") .. "}"
    end
  end

  -- Userdata, classes implemented in C++ with json tostring methods.
  if value_type == "userdata" then
    -- We may be able to serialize to json using the class tostring method.
    local str = base.tostring(value)
    if str and #str > 1 and "{" == string.sub(str, 0, 1) then
      return str
    end
  end

  -- Unknown. Null string.
  return "null"
end

function node.json.command_to_json(command, ...)
  return
    "{\"result\": " ..
    node.system.command_to_string(node.json.result_to_json, command, ...) ..
    ", \"code\": 0}"
end

function node.json.command_to_jsonrpc(command, ...)
  return
    node.system.command_to_string(node.json.result_to_json, command, ...)
end

function node.json.result_to_json(...)
  local result = {}

  local value
  for _,value in base.ipairs({...}) do
    table.insert(result, value)
  end

  return node.json.encode(result, true)
end

function node.ui.calibrate_node(id, calibrate_accel, calibrate_gyro, calibrate_factory)
  if not id then
    id = ""
  end

  if not calibrate_accel then
    calibrate_accel = false
  end

  if not calibrate_gyro then
    calibrate_gyro = false
  end

  if not calibrate_factory then
    calibrate_factory = false
  end

  if not node.is_configured(id) then
    return false, "failed precondition \"is_configured(\"" .. id .. "\") == true\""
  end

  local should_start = node.is_reading()
  node.close()

  local result, message
  if calibrate_gyro then
    local list = node.configuration()
    if not list then
      return false, "failed precondition \"configuration() ~= nil\""
    end

    local item
    for _,item in pairs(list) do
      if item.active and not item.bus and 'LFP' ~= item.id and 'RFP' ~= item.id then
        result, message = node.system.calibrate_from_take(
          item.id, calibrate_accel, calibrate_gyro)
        if not result then
          return result, message
        end

        if calibrate_factory then
          local source_type = string.match(item.source, "([^:]+)://")
          if source_type == 'usb' or source_type == 'serial' then
            result, message = node.write_configuration(item.id)
            if not result then
              return result, message
            end

            if should_start and source_type == 'usb' then
              node.system.sleep(1)
            end
          else
            return false, "failed to write device memory, requires USB connection"
          end
        end
      end
    end
  else
    result, message = node.system.calibrate_from_take(
      id, calibrate_accel, calibrate_gyro)
    if not result then
      return result, message
    end

    if calibrate_factory then
      local item = node.get_node_by_id(id)
      if not item then
        return false, "failed to find " .. _NAME .. " with id \"" .. id .. "\""
      end

      local source_type = string.match(item.source, "([^:]+)://")
      if source_type == 'usb' or source_type == 'serial' then
        result, message = node.write_configuration(id)
        if not result then
          return result, message
        end

        if should_start and source_type == 'usb' then
          node.system.sleep(1)
        end
      else
        return false, "failed to write device memory, requires USB connection"
      end
    end
  end

  if should_start then
    return node.start()
  else
    return true, ""
  end
end

function node.ui.create_rest_pose(id)
  if id then
    local item = node.get_node_by_id(id)
    if not item then
      -- Second, if not available by id, then try the name field.
      item = node.get_node("name", id)
      if item then
        id = configured_node.id
      end
    end

    if item and item.parent then
      id = item.parent
    end

    return node.set_pose_marker(id)
  else
    return node.set_pose_marker()
  end
end

function node.ui.get_pipeline_select(id)
  local result = -1

  if id then
    local item = node.get_node_by_id(id)
    if item then
      result = item.pipeline_select
    end
  else
    local list = node.configuration()
    if list then
      local item
      for _,item in base.pairs(list) do
        if -1 == result then
          result = item.pipeline_select
        elseif item.pipeline_select ~= result then
          result = 0
          break
        end
      end
    end
  end

  return result
end

function node.ui.reset()
  return node.auto_connect(true)
end

function node.ui.reset_position(id)
  if id then
    local item = node.get_node_by_id(id)
    if not item then
      -- Second, if not available by id, then try the name field.
      item = node.get_node("name", id)
      if item then
        id = item.id
      end
    end

    if item and item.parent then
      id = item.parent
    end

    return node.set_pose(id)
  else
    return node.set_pose()
  end
end

function node.ui.start(id)
  if not node.is_configured(id) then
    local result, message = node.scan()
    if not result then
      return result, message
    end
  end

  return node.start(id)
end

function node.ui.toggle_reading(id)
  if node.is_reading(id) then
    return node.close(id)
  else
    local result, message = node.start(id)
    if not result then
      node.close(id)
      return result, message
    end
  end

  return true, ""
end

function node.ui.toggle_inertial(id)
  local value = node.ui.get_pipeline_select(id)
  if 5 == value then
    value = 0
  else
    value = 5
  end

  local result, message = node.set_pipeline_select(value, id)
  if result then
    -- This changed state that we care about in the UI. Trigger a monitor
    -- update. State code 15 denotes a change to the pipeline_select field.
    node.system.trigger_state_change(15, base.tostring(value))
  end

  return result, message
end

function node.ui.get_auto_export_filename()
  local filename = node.system.get_default("_auto_export_filename")
  if "" ~= filename then
    return true, node.system.file_canonical(filename)
  else
    return false, ""
  end
end

function node.ui.enable_hybrid(name)
  if node.is_connected() then
    return false, "failed precondition \"is_connected() == false\""
  end

  if not name then
    name = ""
  end

  local item
  for _,item in base.ipairs({"shadow_openvr", "shadow_natnet"}) do
    local is_enabled = math.floor(node.system.get_default(item))
    if name == item then
      if is_enabled == 0 then
        node.system.set_default(item, 1)
        node.system.load_plugin(item)
      end
    else
      if is_enabled ~= 0 then
        node.system.unload_plugin(item)
        node.system.set_default(item, 0)
      end
    end
  end

  node.system.set_default("hybrid_plugin", name)

  return true, ""
end

--
-- End internal function definitions.
--

--
-- Return module style table of all public node.* definitions.
--
return node
