--
-- api/initialize.lua
-- 
-- Return a Lua string that is used to setup and run the Shadow services. Used
-- at install time to generate the user specific "Motion/default/initialize.lua"
-- script.
--
local pref = node.system.get_preference()
if not pref then
  error("no system preferences found")
end

local data_path = pref.data_path
if not data_path then
  data_path = "data"
end

local log_path = pref.log_path
if not log_path then
  log_path = "default/log.json"
end

print([[--
-- The "bootstrap.lua" file calls this script at start up. Define all
-- site specific preferences in this file.
--
-- Generated by the "api/initialize.lua" script.
--
-- @file    initialize.lua
-- @version ]] .. node._VERSION .. [[

--

--
--
-- Check if the system is already initialized. If it is simply
-- return now.
--
if node.system.is_initialized() then
  return
end

local result
local message

--
-- Set the output (and search) path for writable data.
-- Examples:
--  "data", "c:/data", "//network_name/data", or "H:\\Take Data"
--
local result, message = node.system.set_data_path("]] .. data_path .. [[")
if not result then
  error(message)
end

--
-- Set the system wide log file name.
--
result, message = node.system.log("]] .. log_path .. [[")
if not result then
  error(message)
end

--
-- Open default database file.
--
if node.system.file_exists("default/database.db") then
  node.system.open_database("default/database.db")
end

--
-- Load default license file.
--
if node.system.file_exists(node._license_filename) then
  node.open(node._license_filename)
end

--
-- Load all plugins in the search path.
--
node.system.load_plugin()

--
-- Define and start services.
--
local service = {
  -- Switchboard for monitor and other data stream services.
  ["monitor"] = 32074,
  -- Interactive console service, optional.
  ["console"] = 32075,
  -- Provide real-time access to all data streams. The client selects channels
  -- at the start of the connection.
  ["device_configurable"] = 32076,
  -- Accept HTTP connections. Required for the web based user interface.
  ["http"] = 32080
}

--
-- If we failed to start any of the requested services
-- call it a fatal error.
--
result, message = node.system.start_services(service)
if not result then
  error(message)
end

--
-- Set the current geographic location. This is required to compute the
-- geomagnetic field reference value. Note that the "location.mGeocode" file
-- will override the location entered in this file.
--
-- Location address.
-- ]] .. pref.location.address .. [[

--
-- Decimal degrees (-90, 90) where negative latitude denotes Southern hemisphere
local latitude = ]] .. pref.location.latitude .. [[

-- Decimal degrees (-180, 180) where negative longitude denotes Western hemisphere
local longitude = ]] .. pref.location.longitude .. [[

-- Elevation from surface of earth in Meters, [-1000, 600000]
local elevation = ]] .. pref.location.elevation .. [[

-- Street address. Just informative.
local street_address = "]] .. pref.location.address .. [["

--
-- Load default location file.
--
if node.system.file_exists(node._location_filename) and
   node.open(node._location_filename) then
else
  -- Check if we already have a geographic location loaded from the database or
  -- other command.
  local preference = node.system.get_preference()
  if not preference or 0 == preference.location.active then
    result, message = node.system.location(
      latitude, longitude, elevation, street_address)
    if not result then
      error(message)
    end
  end
end

--
-- Load default ractor file.
--
if node.system.file_exists(node._ractor_filename) then
  node.open(node._ractor_filename)
end

--
-- Load default configuration file.
--
if node.system.file_exists(node._configuration_filename) then
  node.open(node._configuration_filename)
end

--
-- Scan for available devices if the configuration is empty.
--
if not node.is_configured() then
  node.scan()
end

--
-- Run user start up script.
--
if node.system.file_exists("default/user.lua") then
  node.open("default/user.lua")
end

--
-- Start reading from all configured devices.
--
if node.is_configured() then
  node.start()
end

--
-- End initialize.lua
--]])
