--
-- api/node_to_mel.lua
--
-- Output a MEL script (for Autodesk Maya) that creates a skeleton and
-- channel name mapping that matches to current configuration. Used to
-- dynamically generate a Maya skeleton for live streaming.
--

local list = node.configuration()

print([[
/**
  Automatically generated by the "node_to_mel.lua" script. Input is a node
  configuration. The output is a Maya MEL command set to create the skeleton,
  effector locators, and preview channel associations.

  Example usage (in MEL):

  // Create joint hierarchy.
  mDevice_create_skeleton();

  // Create locators for each of the joint FK/IK effectors.
  mDevice_create_marker();

  // Preview real-time data. Uses the Maya mocap device with associations
  // defined to the skeleton and locator set set we just loaded.
  mDevice_device(1);
*/

// Version string of the mDevice library. Also makes sure that the library is
// loaded and ready to be used.
string $mDevice_name = mDevice();

global proc string mDevice_group_name()
{
  return "ConfigurationGroup";
}

global proc string mDevice_create_skeleton()
{
  global int $MDEVICE_IMPORT_MASS;

  string $group = mDevice_group_name();
  select -clear;
  if (objExists($group)) {
    error("Object named \"" + $group + "\" already exists");
  }
  group -empty -world -name $group;
]])

local node_visit_order = node.get_node_visit_order()
local node_path_list = {}
local node_key_list = {}
if node_visit_order then
  local idx, id
  for idx,id in ipairs(node_visit_order) do
    local path = ''

    local ancestor = node.get_node_ancestor(id)

    local ancestor_id
    for _,ancestor_id in pairs(ancestor) do
      path = path .. '|' .. ancestor_id
    end

    local conf_node = node.get_node_by_id(id)

    node_path_list[idx] = path
    node_key_list[idx] = conf_node.key

    if #ancestor == 0 then
      print([[
  {
    select -replace ($group);
    group -empty -parent `ls -selection -long -type transform` -name "]] ..
    id .. [[";
  }
]])
    else
      local com_translate = table.concat(conf_node.marker, ' ', 5, 7)

      print([[
  {
    select -replace ($group + "]] .. path .. [[");
    string $name = `joint -radius 3 -relative -orientation ]] ..
    table.concat(conf_node.rotate, ' ') .. [[ -position ]] .. table.concat(conf_node.translate, ' ') ..
    [[ -name "]] .. id .. [["`;
]])
      if #ancestor == 1 then
        print([[
    setAttr ($name + ".type") 1;
]])
      end
print([[
    if ($MDEVICE_IMPORT_MASS) {
      string $locator_name[] = `spaceLocator -name ($name + "_COM")`;
      setAttr ($locator_name[0] + ".translate") ]] .. com_translate .. [[;
      addAttr -attributeType double -defaultValue 0 -longName mass $locator_name[0];
      setAttr ($locator_name[0] + ".mass") ]] .. conf_node.mass .. [[;
      parent -relative $locator_name[0] $name;
    }
  }
]])
    end
  end
end

print([[
  return $group;
}
]])

print([[
global proc string[] mDevice_device_channel_list()
{
  string $deviceChannelList[] = {]])
  if node_visit_order then
    local tbl = {}

    local id
    for _,id in ipairs(node_visit_order) do
      table.insert(tbl, '"' .. id .. '"')
    end
    print('    ' .. table.concat(tbl, ',\n    '))
  end
print([[
  };
  return $deviceChannelList;
}
]])

print([[
global proc string[] mDevice_object_channel_list()
{
  string $group = mDevice_group_name();
  if (!objExists($group)) {
    string $list[] = `ls -selection -type transform`;
    if (1 == size($list)) {
      $group = $list[0];
    }
  }
  if (!objExists($group)) {
    $group = "";
  } else {
    $group = $group + "|";
  }

  string $objectChannelList[] = {]])
  if node_visit_order then
    local tbl = {}

    local idx, id
    for idx,id in ipairs(node_visit_order) do
      local full_path = string.gsub(node_path_list[idx] .. '|' .. id, '^|', '', 1)
      table.insert(tbl, '"' .. full_path .. '"')
    end

    print('    ' .. table.concat(tbl, ',\n    '))
  end
print([[
  };
  return $objectChannelList;
}
]])

print([[
global proc int[] mDevice_device_key_list()
{
  int $deviceKeyList[] = {]])
  if node_visit_order then
    local tbl = {}

    local idx, id
    for idx,id in ipairs(node_visit_order) do
      table.insert(tbl, node_key_list[idx])
    end
    print('    ' .. table.concat(tbl, ',\n    '))
  end
print([[
  };
  return $deviceKeyList;
}
]])

print([[
global proc string mDevice_create_marker()
{
  string $marker_name = mDevice_marker_name();
  string $parent = mDevice_group_name();
  string $group = $parent + "|" + $marker_name;

  select -clear;
  if (objExists($group)) {
    error("Object named \"" + $group + "\" already exists");
  }
  group -empty -parent $parent -name $marker_name;

  for ($item_name in mDevice_device_channel_list()) {
    string $locator_name[] = `spaceLocator -name ($item_name + "_" + $marker_name)`;
    addAttr
      -attributeType double -defaultValue 0
      -shortName w -longName weight -keyable true $locator_name[0];
    parent $locator_name[0] $group;
  }

  return $group;
}

global proc string mDevice_create_character()
{
  string $parent = mDevice_group_name();
  string $group = "TakeCharacter";

  if (!objExists($parent)) {
    error("Object named \"" + $parent + "\" does not exist");
  }

  if (objExists($group)) {
    error("Object named \"" + $group + "\" already exists");
  }

  select -replace -hierarchy $parent;

  $group = `character
    -name $group
    -excludeScale
    -excludeVisibility`;

  select -clear;

  return $group;
}

global proc mDevice_rest_pose()
{
  string $group = mDevice_group_name();
  {
    string $object = ($group + "|Body|Hips");

    setAttr ($object + ".translate") 0 100 0;
    select $object;
    select -hi;
    makeIdentity -apply false -t 0 -r 1 -s 0;
    select -clear;
  }
}

global proc string[] mDevice_device_attribute_list()
{
  string $deviceAttributeList[3] = {
    "rotateXYZ:X", "rotateXYZ:Y", "rotateXYZ:Z"
  };

  return $deviceAttributeList;
}

global proc string[] mDevice_object_attribute_list()
{
  string $objectAttributeList[3] = {
    "rotateX", "rotateY", "rotateZ"
  };

  return $objectAttributeList;
}
]])