--
-- Return a string JSON description of the user settings on the Motion Service.
-- Show the active location, the current ractor, and any system preferences that
-- we may want to view.
--
-- @file    json/settings.lua
-- @version 4.0
--

local list, item

local version = node.get_version()

local settings = {
  ['type'] = 'settings',
  ['location'] = {},
  ['ractor'] = {},
  ['system'] = {
    ['auto_export'] = 0,
    ['auto_export_path'] = '{path}/take.fbx',
    ['auto_track_take'] = 1,
    ['export_type'] = 'fbx',
    ['floor_plane'] = 1,
    ['hybrid_plugin'] = '',
    ['rest_pose'] = 0,
    ['use_calibration_db'] = 1
  },
  ['export_types'] = {},
  ['imu_channel_masks'] = {
    {['key'] = 0, ['name'] = 'Default'},
    -- A, M, G, p
    {['key'] = 16805888, ['name'] = 'Raw'},
    -- Gq, Gdq, la, em, p
    {['key'] = 33030163, ['name'] = 'Quat'},
    -- Raw and Quat
    {['key'] = 16805888 + 33030163, ['name'] = 'Raw-Quat'}
  },
  ['version'] = version
}

if version.is_bus or version.is_shadow then
  local Measurement = 1 << 15
  local MeasurementRaw = Measurement | 1 << 16
  local MeasurementTether = Measurement | 1 << 17
  local MeasurementRate = Measurement | 1 << 18

  -- For the remote Bus or USB tethering to a computer. Show the data
  -- acquisition and internal IMU update rate settings.
  if version.is_bus then
    if version.capability_mask & MeasurementRaw ~= 0 then
      settings['system']['imu_channel_mask'] = 0
    end

    if version.capability_mask & MeasurementRate ~= 0 then
      settings['system']['imu_time_step'] = 0.01

      settings['sample_rate'] = {
        {['key'] = 50, ['name'] = '50'},
        {['key'] = 100, ['name'] = '100'},
        {['key'] = 200, ['name'] = '200'},
        {['key'] = 400, ['name'] = '400'},
        {['key'] = 500, ['name'] = '500'},
        {['key'] = 800, ['name'] = '800'},
        {['key'] = 1000, ['name'] = '1000'}
      }
    end
  end

  settings['shadow_conf'] = {}
  settings['system']['shadow_conf'] = ''

  list = node.system.get_system_conf_list()
  if list then
    for _,item in pairs(list) do
      table.insert(settings.shadow_conf, copy_usertype(item))
    end
  else
    settings.shadow_conf = {
      {['name'] = '', ['value'] = 'Default', ['description'] = ''},
      {['name'] = 'hybrid', ['value'] = 'Shadow Hybrid', ['description'] = ''},
      {['name'] = 'biomech', ['value'] = 'Shadow Biomech', ['description'] = ''},
      {['name'] = 'v3', ['value'] = 'Shadow v3', ['description'] = ''},
      {['name'] = 'v2', ['value'] = 'Shadow v2', ['description'] = ''}
    }
  end

  if version.is_shadow and not version.is_bus then
    settings['hybrid_plugin'] = {
      {['name'] = '', ['value'] = 'None', ['description'] = ''},
      {['name'] = 'shadow_openvr', ['value'] = 'Vive Tracker and SteamVR', ['description'] = ''}
    }

    if version.platform_name == 'Windows' then
      table.insert(settings.hybrid_plugin, {
        ['name'] = 'shadow_natnet',
        ['value'] = 'OptiTrack and Motive',
        ['description'] = ''
      })
    end
  end

  settings['shadow_pose'] = {}
  settings['system']['shadow_definition'] = 'default/definition.xml'

  settings['shadow_pose'] = {
    {['name'] = 'default/definition.xml', ['value'] = 'T-Pose', ['description'] = ''},
    {['name'] = 'default/definition-apose.xml', ['value'] = 'A-Pose', ['description'] = ''}
  }
end

local location_to_table = function(obj)
  return {
    ['key'] = obj['key'],
    ['name'] = obj['name'],
    ['active'] = obj['active'],
    ['canonical_address'] = obj['canonical_address'],
    ['latitude'] = obj['latitude'],
    ['longitude'] = obj['longitude'],
    ['elevation'] = obj['elevation']
  }
end

local ractor_to_table = function(obj)
  return {
    ['name'] = obj['name'],
    ['active'] = obj['active'],
    ['key'] = obj['key'],
    ['measurement'] = {
      ['unit'] = 'cm',
      ['height'] = obj.height,
      ['arm'] = obj.arm,
      ['leg'] = obj.leg
    }
  }
end

-- Location.
list = node.system.get_location_list()
if list then
  for _,item in pairs(list) do
    table.insert(settings.location, copy_usertype(item))
  end
else
  local preference = node.system.get_preference()
  if preference then
    table.insert(settings.location, copy_usertype(preference.location))
  end
end

-- Ractor.
list = node.system.get_ractor_list()
if list then
  for _,item in pairs(list) do
    table.insert(settings.ractor, copy_usertype(item))
  end
end

-- System preferences.
local key, value
for key,value in pairs(settings.system) do
  sv = node.system.get_default(key)
  if sv ~= "" then
    settings.system[key] = sv

    -- Cast to number format.
    if type(value) == "number" then
      settings.system[key] = tonumber(settings.system[key])
    end
  end
end

-- List of available export types.
list = node.system.export_type()
if list then
  for _,item in pairs(list) do
    table.insert(settings.export_types, copy_usertype(item))
  end
end

return node.json.encode(settings)
